import logging
from itertools import islice
from .client import annotate_variants_list
from typing import Optional
from tqdm import tqdm


try:
    # try with a fast c-implementation ...
    import cyvcf2 as cyvcf2
except ImportError:
    # ... otherwise fallback python implementation
    logging.error(
        'No cyvcf2 module installed. Install, f.ex. by executing "pip install cyvcf2".'
    )


def annotate_vcf(
    input_vcf_path: str,
    output_vcf_path: str,
    genome: str = "hg38",
    use_ensembl: bool = True,
    use_refseq: bool = True,
    flatten_consequences: bool = True,
    batch_size: int = 500,
    username: Optional[str] = None,
    api_key: Optional[str] = None,
    use_netrc: bool = True,
    endpoint_url: str = "https://api.genebe.net/cloud/api-public/v1/variants",
):
    """
    Annotates a VCF file using the annotate_variants_list function.

    Args:
        input_vcf_path (str): Path to the input VCF file.
        output_vcf_path (str): Path to the output annotated VCF file.
        genome (str, optional): The genome version for annotation (e.g., 'hg38').
            Defaults to 'hg38'.
        use_ensembl (bool, optional): Whether to use Ensembl for annotation.
            Defaults to True.
        use_refseq (bool, optional): Whether to use RefSeq for annotation.
            Defaults to True.
        flatten_consequences (bool, optional): If set to False, return consequences as a list of objects.
            If set to True, only the most important consequence is returned in a flat form.
            Defaults to True.
        batch_size (int, optional): The size of each batch for processing variants.
            Defaults to 100. Must be smaller or equal 1000.
        username (str, optional): The username for authentication.
            Defaults to None.
        api_key (str, optional): The API key for authentication.
            Defaults to None.
        use_netrc (bool, optional): Whether to use credentials from the user's
            .netrc file for authentication. Defaults to True.
        endpoint_url (str, optional): The API endpoint for variant annotation.
            Defaults to 'https://api.genebe.net/cloud/api-public/v1/variants'.
    """

    print(f"Annotating VCF file: {input_vcf_path}")
    print(f"Using genome: {genome}")
    print(f"Batch size: {batch_size}")
    print(f"Output file: {output_vcf_path}")

    # Open VCF file
    vcf_reader = cyvcf2.VCF(input_vcf_path)

    # Add header to the VCF file
    vcf_reader.add_info_to_header(
        {
            "ID": "acmg_score",
            "Number": "1",
            "Type": "Integer",
            "Description": "ACMG Score",
        }
    )

    # Open the output VCF file for writing
    vcf_writer = cyvcf2.Writer(output_vcf_path, vcf_reader)

    # Batch size for reading variants
    batch_size = 1000

    with tqdm() as pbar:
        # Iterate over variants in batches
        while True:
            # Read a batch of variants
            batch = list(islice(vcf_reader, batch_size))
            if not batch:
                break  # Break if there are no more variants

            # Process the batch
            variants_batch = [
                f"{variant.CHROM}-{variant.POS}-{variant.REF}-{variant.ALT[0]}"
                for variant in batch
            ]
            annotated_variants = annotate_variants_list(
                variants_batch,
                genome=genome,
                use_ensembl=use_ensembl,
                use_refseq=use_refseq,
                flatten_consequences=flatten_consequences,
                batch_size=batch_size,
                username=username,
                api_key=api_key,
                use_netrc=use_netrc,
                endpoint_url=endpoint_url,
                progress_bar=False,
            )

            # Add ACMG scores to the variants and write to the output VCF file
            for variant, annotated_variant in zip(batch, annotated_variants):
                acmg_score = annotated_variant.get("acmg_score")
                if acmg_score is not None:
                    variant.INFO["acmg_score"] = acmg_score
                else:
                    logging.warning(
                        f"No annotation for {variant.CHROM}-{variant.POS}-{variant.REF}-{variant.ALT[0]}"
                    )

                vcf_writer.write_record(variant)
            pbar.update(len(batch))

    # Close the VCF writer
    vcf_writer.close()
